# coding: utf-8
"""

    GaussianProcess.py

    Copyright (c) 2016, Masatsuyo Takahashi, KEK-PF
"""
import pymc.gp, sklearn
import numpy                as np
from sklearn                import gaussian_process
from scipy.interpolate      import UnivariateSpline

"""
    
    https://healthyalgorithms.com/2015/01/20/non-parametric-regression-in-python-gaussian-processes-in-sklearn-with-a-little-pymc/
"""
def iso_matern(theta, d):
    """
    Matern correlation model.

        theta, d --> r(diff_degree=theta[0], amp=theta[1], scale=theta[2]; d)

    Parameters
    ----------
    theta : array_like
        An array with shape 3 (isotropic) giving the
        autocorrelation parameters (diff_degree, amp, scale)
    d : array_like
        An array with shape (n_eval, n_features) giving the componentwise
        distances between locations x and x' at which the correlation model
        should be evaluated.
    Returns
    -------
    r : array_like
        An array with shape (n_eval, ) with the values of the autocorrelation
        model.
    """
    # print( theta )
    #return sklearn.gaussian_process.correlation_models.generalized_exponential(theta[:2], d)
    theta = np.asarray(theta, dtype=np.float).reshape(3)
    #theta[0] = 2
    #theta[1] = .4
    d = np.asarray(d, dtype=np.float)

    assert theta.shape == (3,), 'expected theta.shape == (3,),  with diff_degree=theta[0], amp=theta[1], scale=theta[2]'

    r = pymc.gp.cov_funs.matern.euclidean(
        d, np.array([0]),
        diff_degree=theta[0], amp=theta[1], scale=theta[2])
    return np.array(r).reshape(d.shape[0],)


class Smoother:
    def __init__( self, x, y, is_guinier=True, head_curvature=False, nugget=1e-3, is_good_quality=True ):
        X = np.atleast_2d( x ).T
        self.gp = gaussian_process.GaussianProcess(corr=iso_matern,
                    theta0=[1.5,1,10], nugget=nugget )
                    # theta0=[ 3, 1, 10 ], nugget=nugget )
        self.gp.fit( X, y )
        self.is_guinier = is_guinier

        if not head_curvature: return

        sy = self.predict( x )
        spline = UnivariateSpline( x, sy )
        yˈ  = spline.derivative(1)( x )
        yˈˈ = spline.derivative(2)( x )
        cy = yˈˈ / np.power( 1 + yˈ**2, 3/2 )
        self.yˈ  = yˈ
        self.yˈˈ = yˈˈ
        self.cy  = cy
        self.cv_array   = np.percentile( cy, [ 5, 20, 50, 80, 95 ] )
        print( 'cv_array=', self.cv_array )

    def __call__( self, x ):
        if np.isscalar( x ):
            assert( False )     # not implemented yet
        else:
            return self.predict( x )

    def predict( self, x ):
        x = np.atleast_2d( x ).T
        y_pred, sigma2_pred = self.gp.predict(x, eval_MSE=True)
        self.sigma_pred = np.sqrt(sigma2_pred)
        return y_pred


class SmootherChain:
    def __init__( self, x, y, is_guinier=True, boundaries=None, nuggets=None, is_good_quality=True ):
        self.x  = x
        self.y  = y
        self.is_guinier = is_guinier

        size    = x.shape[0]

        if boundaries is None:
            quarter = size // 4
            half    = quarter*2
            boundaries = [ 0, quarter, half, size ]

        self.intervals  = []
        for i, f in enumerate( boundaries[:-1] ):
            t = boundaries[ i+1 ]
            nugget = 1e-3 if nuggets is None else nuggets[i]
            slice_ = slice( f, t )
            x_ = x[slice_]
            y_ = y[slice_]
            smoother = Smoother( x_, y_, nugget=nugget )

            if i == 0 and f > 0:
                self.intervals.append( [ 0, f, smoother ] )

            self.intervals.append( [ f, t, smoother ] )

        if t < size:
            self.intervals.append( [ t, size, smoother ] )

    def __call__( self, x ):
        if np.isscalar( x ):
            return self.predict_scalar( x )
        else:
            return self.predict( x )

    def predict( self, x ):
        y_array = []

        # print( 'predict: x=', x )

        i = 0
        for f, t, smoother in self.intervals:
            xf = 0.0        if i == 0               else self.x[f]
            xt = self.x[t]  if t < self.x.shape[0]  else np.inf
            x_ = x[ np.logical_and( x >= xf, x < xt ) ]
            # print( 'f,t=', [f, t], '; xf, xt=', [ xf, xt ], '; x_.shape[0]=', x_.shape[0] )
            if x_.shape[0] > 0:
                y_array.append( smoother.predict( x_ ) )
            i += 1

        return np.hstack( y_array )

    def predict_scalar( self, x ):
        i = 0
        for f, t, smoother in self.intervals:
            if i == 0 and x < self.x[f]:
                return smoother.predict( np.array( [x] ) )[0]
            if x >= self.x[f] and x <= self.x[t-1]:
                return smoother.predict( np.array( [x] ) )[0]

            i += 1

        return smoother.predict( np.array( [x] ) )[0]
