"""
    McGaussian.py

    original example:
        https://pymc-devs.github.io/pymc3/notebooks/GP-smoothing.html

    Copyright (c) 2016, Masatsuyo Takahashi, KEK-PF
"""
import numpy as np
import scipy.stats as stats
import matplotlib.pyplot    as plt
import pymc3 as pm
from theano import shared
from pymc3.distributions.timeseries import GaussianRandomWalk
from scipy import optimize
from scipy.interpolate      import UnivariateSpline

LARGE_NUMBER = 1e5

class Smoother:
    def __init__( self, x, y, smoothing=0.95 ):
        self.x  = x
        self.y  = y

        self.model = pm.Model()
        with self.model:
            self.smoothing_param = smoothing_param = shared(0.9)
            mu = pm.Normal("mu", sd=LARGE_NUMBER)
            tau = pm.Exponential("tau", 1.0/LARGE_NUMBER)
            self.z = z = GaussianRandomWalk("z",
                                   mu=mu,
                                   tau=tau / (1.0 - smoothing_param),
                                   shape=y.shape)
            obs = pm.Normal("obs",
                            mu=z,
                            tau=tau / smoothing_param,
                            observed=y)

        z_val = self.infer_z( smoothing )
        self.smoother = UnivariateSpline( x, z_val, k=5 )

    def infer_z( self, smoothing ):
        with self.model:
            self.smoothing_param.set_value(smoothing)
            res = pm.find_MAP(vars=[self.z], fmin=optimize.fmin_l_bfgs_b)
            return res['z']

    def __call__( self, x ):
        return self.smoother( x )

if __name__ == '__main__':
    import matplotlib.pyplot    as plt

    x = np.linspace(0, 50, 100)
    y = (np.exp(1.0 + np.power(x, 0.5) - np.exp(x/15.0)) +
         np.random.normal(scale=1.0, size=x.shape))

    smoothing = 0.95
    smoother = Smoother( x, y, smoothing )
    z_val = smoother( x )

    plt.plot(x, y);
    plt.xlabel("x");
    plt.ylabel("y");

    lin = stats.linregress(x, y)
    plt.plot(x, lin.intercept + lin.slope * x);

    plt.plot(x, z_val);

    plt.title("Smoothing={}".format(smoothing));

    plt.show()
